@@ -1,12 +1,12 @@
 """Provides access to config variables."""
 
 import os
-from typing import Optional, List, Dict
+from typing import Optional, List, Dict, Any
 
 import toml
-from cloudpathlib import AnyPath
 from frozendict import frozendict
 
+from cpg_utils import to_path
 
 # We use these globals for lazy initialization, but pylint doesn't like that.
 # pylint: disable=global-statement, invalid-name
@@ -14,19 +14,35 @@ _config_paths = _val.split(',') if (_val := os.getenv('CPG_CONFIG_PATH')) else [
 _config: Optional[frozendict] = None  # Cached config, initialized lazily.
 
 
-def _validate_configs(config_paths: List[str]) -> None:
+def _validate_configs(config_paths: list[str]) -> None:
     if [p for p in config_paths if not p.endswith('.toml')]:
         raise ValueError(
             f'All config files must have ".toml" extensions, got: {config_paths}'
         )
-    if bad_files := [p for p in config_paths if not AnyPath(p).exists()]:
-        raise ValueError(f'Some config files do not exist: {bad_files}')
+
+    paths = [to_path(p) for p in config_paths]
+    if bad_paths := [p for p in paths if not p.exists()]:
+        raise ValueError(f'Some config files do not exist: {bad_paths}')
+
+    # Reading each file to validate syntax:
+    exception_by_path = {}
+    for p in paths:
+        with p.open() as f:
+            try:
+                toml.loads(f.read())
+            except toml.decoder.TomlDecodeError as e:
+                exception_by_path[p] = e
+    if exception_by_path:
+        msg = f'Failed parsing some config files:'
+        for path, exception in exception_by_path.items():
+            msg += f'\n\t{path}: {exception}'
+        raise ValueError(msg)
 
 
 _validate_configs(_config_paths)
 
 
-def set_config_paths(config_paths: List[str]) -> None:
+def set_config_paths(config_paths: list[str]) -> None:
     """Sets the config paths that are used by subsequent calls to get_config.
 
     If this isn't called, the value of the CPG_CONFIG_PATH environment variable is used
@@ -45,6 +61,29 @@ def set_config_paths(config_paths: List[str]) -> None:
         _config = None  # Make sure the config gets reloaded.
 
 
+def prepend_config_paths(config_paths: list[str]) -> None:
+    """
+    Prepend to the list of config paths. Equivalent to `dict.set_defaults`: any
+    values in current CPG_CONFIG_PATH will have the precedence over the provided
+    `config_paths` when merging the configs.
+    """
+    if _env_var := os.environ.get('CPG_CONFIG_PATH'):
+        config_paths.extend(_env_var.split(','))
+
+    set_config_paths(config_paths)
+
+
+def append_config_paths(config_paths: list[str]) -> None:
+    """
+    Append to the list of config paths. Any values in new configs will have the
+    precedence over the existing CPG_CONFIG_PATH when merging the configs.
+    """
+    if _env_var := os.environ.get('CPG_CONFIG_PATH'):
+        config_paths = _env_var.split(',') + config_paths
+
+    set_config_paths(config_paths)
+
+
 def get_config() -> frozendict:
     """Returns the configuration dictionary.
 
@@ -111,7 +150,7 @@ def read_configs(config_paths: List[str]) -> frozendict:
 
     config: dict = {}
     for path in config_paths:
-        with AnyPath(path).open() as f:
+        with to_path(path).open() as f:
             config_str = f.read()
             update_dict(config, toml.loads(config_str))
     return frozendict(config)
@@ -125,3 +164,27 @@ def update_dict(d1: Dict, d2: Dict) -> None:
             update_dict(v1, v2)
         else:
             d1[k] = v2
+
+
+class ConfigError(Exception):
+    """
+    Error retrieving keys from config.
+    """
+
+
+def retrieve(key: list[str] | str, default: Any | None = None) -> str:
+    """
+    Retrieve key from config, assuming nested key specified as a list of strings.
+    """
+    if isinstance(key, str):
+        key = [key]
+
+    d = get_config()
+    for k in key[:-1]:
+        if k not in d:
+            raise ConfigError(f'Key "{k}" not found in {d}')
+        d = d[k]
+    k = key[-1]
+    if k not in d and not default:
+        raise ConfigError(f'Key "{k}" not found in {d}')
+    return d.get(k, default)
